/*
 * Copyright (C) 1996-2003 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"
__RCSID("$LAAS: GDHE_xdr.c,v 1.7 2003/05/05 15:22:10 matthieu Exp $");

/******************************************************\
**                     xdr.c                          **
**                    -------                         **
**  This file implements all functions necessary for  **
** XDR communication with the packets.h file.         **
**                                                    **
**  There is one translation routine for each object  **
** type, namely:                                      **
**        xdr_location                                **
**        xdr_segment                                 **
**        xdr_position                                **
**        xdr_v_segment                               **
**        xdr_v_location                              **
**        xdr_location                                **
**        xdr_trajectory                              **
**        xdr_polygon                                 **
**  (see file packets.h for info on objects)          **
**                                                    **
**  The file xdr.c is to be compiled and archived as  **
** a user-library.                                    **
**                                                    **
**      (C) Je'ro^me Perret, 1992                     **
\******************************************************/
/**====================================================
  Modifications:
      03/05/1994   S.Suzuki          add 'select_robot_type()' 
      10/06/1994   Suzuki & Perret   adaptation for VxWorks
  ====================================================**/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifndef VXWORKS
#include <sys/time.h>
#else
#include <sys/times.h>
#endif
#include <rpc/rpc.h>

#ifdef VXWORKS
#include "vxWorks.h"
#endif

#define  GDHE_CLIENT_VAR
#include "GDHE.h"
#include "GDHE_xdr_prot.h"
#ifdef HAVE_GDHE_TERRAIN
#include "gdhe_terrain.h"
#endif


/* Translation procedures: based on elementary translation *\
** calls by xdr, and nested in one another. Careful: don't **
** call xdr_location directly inside xdr_position for      **
** example, because upon decoding no free space is yet     **
** allocated. Let xdr_reference do the job. For the same   **
** reason, don't call xdr_packet without setting pointer   **
\* NULL.                                                   */

int xdr_location ( XDR* xdrsp, location* locp )
{
  if ( !xdr_float ( xdrsp, &(locp->x) ) )
    return FALSE;
  if ( !xdr_float ( xdrsp, &(locp->y) ) )
    return FALSE;
  return TRUE;
}

int xdr_segment ( XDR* xdrsp, segment* segp)
{
  if ( !xdr_reference ( xdrsp, (char**) &(segp->p1),
		       sizeof ( location ),
		       (xdrproc_t)xdr_location ) )
    return FALSE;
  if ( !xdr_reference ( xdrsp, (char**) &(segp->p2),
		       sizeof ( location ),
		       (xdrproc_t)xdr_location ) )
    return FALSE;
  return TRUE;
}

int xdr_position ( XDR* xdrsp, position* posp )
{
  if ( !xdr_reference ( xdrsp, (char **) &(posp->p),
		       sizeof ( location ),
		       (xdrproc_t)xdr_location ) )
    return FALSE;
  if ( !xdr_float ( xdrsp, &(posp->theta) ) )
    return FALSE;
  return TRUE;
}

int xdr_v_location ( XDR* xdrsp, v_location* vlocp )
{
  u_int size;

  if ( !xdr_int ( xdrsp, &(vlocp->number) ) )
    return FALSE;
  size = ( u_int ) vlocp->number;
  if ( !xdr_array ( xdrsp, (char**) &(vlocp->array),
		   &size, vlocp->number,
		   sizeof ( location ),
		   (xdrproc_t)xdr_location )
      || ( size != vlocp->number ) )
    return FALSE;
  return TRUE;
}

int xdr_v_position ( XDR* xdrsp, v_position* vposp )
{
  u_int size;

  if ( !xdr_int ( xdrsp, &(vposp->number) ) )
    return FALSE;
  size = vposp->number;
  if ( !xdr_array ( xdrsp, (char**) &(vposp->array),
		   &size, vposp->number,
		   sizeof ( position ),
		   (xdrproc_t)xdr_position )
      || ( size != vposp->number ) )
    return FALSE;
  return TRUE;
}

int xdr_v_segment ( XDR* xdrsp, v_segment* vsegp )
{
  u_int size;

  if ( !xdr_int ( xdrsp, &(vsegp->number) ) )
    return FALSE;
  size = vsegp->number;
  if ( !xdr_array ( xdrsp, (char**) &(vsegp->array),
		   &size, vsegp->number,
		   sizeof ( segment ),
		   (xdrproc_t)xdr_segment )
      || ( size != vsegp->number ) )
    return FALSE;
  return TRUE;
}

int xdr_trajectory ( XDR* xdrsp, trajectory* trajp )
{
  if ( ! xdr_float ( xdrsp, &(trajp->theta0) ) )
    return FALSE;
  if ( ! xdr_float ( xdrsp, &(trajp->theta1) ) )
    return FALSE;
  if ( ! xdr_reference ( xdrsp, ( char** ) &(trajp->vloc),
			sizeof ( v_location ),
			(xdrproc_t)xdr_v_location ) )
    return FALSE;
  return TRUE;
}

int
xdr_rs_segment(XDR *xdrsp, rs_segment *segp)
{
    if (xdr_float(xdrsp, &(segp->dt)) == FALSE) {
	return(FALSE);
    }
    if (xdr_float(xdrsp, &(segp->dl)) == FALSE) {
	return(FALSE);
    }
    return(TRUE);
}

int 
xdr_rs_trajectory(XDR *xdrsp, rs_trajectory *trajp)
{
    u_int size;

    if (xdr_reference(xdrsp, (char **)&(trajp->initpos),
		      sizeof(position), (xdrproc_t)xdr_position) == FALSE) {
	return(FALSE);
    }
    if (xdr_int(xdrsp, &(trajp->number)) == FALSE) {
	return(FALSE);
    }
    size = trajp->number;
    if ((xdr_array(xdrsp, (char **)&(trajp->array), &size, 
	    size, sizeof(rs_segment), (xdrproc_t)xdr_rs_segment) == FALSE)
	|| (size != trajp->number)) {
	return(FALSE);
    }
    return(TRUE);
}
	
int xdr_polygon ( XDR* xdrsp, polygon* polp )
{
  u_int size;

  if ( !xdr_int ( xdrsp, &(polp->number) ) )
    return FALSE;
  size = ( u_int ) polp->number;
  if ( !xdr_array ( xdrsp, (char**) &(polp->polys),
		   &size, polp->number,
		   sizeof ( v_location ),
		   (xdrproc_t)xdr_v_location )
      || ( size != polp->number ) )
    return FALSE;
  if ( !xdr_array ( xdrsp, (char**) &(polp->bool),
		   &size, polp->number,
		   sizeof ( int ),
		   (xdrproc_t)xdr_int )
      || ( size != polp->number ) )
    return FALSE;
  return TRUE;
}


int 
xdr_dependency_cycle(XDR *xdrsp, dependency_cycle *cyclep)
{
    u_int size = (u_int)cyclep->number;

    if (!xdr_int(xdrsp, &(cyclep->number))) {
	return(FALSE);
    }
    
    if (!xdr_array(xdrsp, (char **)&(cyclep->nodes),
		    &size, cyclep->number, 
		    sizeof(int), (xdrproc_t)xdr_int) 
	|| (size != cyclep->number)) {
	return(FALSE);
    }
    return(TRUE);
}

#ifdef HAVE_GDHE_TERRAIN
int 
xdr_mnt_point(XDR *xdrsp, GDHE_MNT_POINT *p)
{
    if (!xdr_short(xdrsp, &(p->z))) {
	return FALSE;
    }
    if (!xdr_u_short(xdrsp, &(p->type))) {
	return FALSE;
    }
    return TRUE;
}

int 
xdr_mnt(XDR *xdrsp, GDHE_MNT *mnt) 
{
    unsigned int size = mnt->hdr.nrows * mnt->hdr.ncols;

    if (!xdr_double(xdrsp, &(mnt->hdr.x0))) {
	return(FALSE);
    }
    if (!xdr_double(xdrsp, &(mnt->hdr.y0))) {
	return(FALSE);
    }
    if (!xdr_double(xdrsp, &(mnt->hdr.z0))) {
	return(FALSE);
    }
    if (!xdr_double(xdrsp, &(mnt->hdr.dx))) {
	return(FALSE);
    }
    if (!xdr_double(xdrsp, &(mnt->hdr.dy))) {
	return(FALSE);
    }
    if (!xdr_double(xdrsp, &(mnt->hdr.dz))) {
	return(FALSE);
    }
    if (!xdr_u_int(xdrsp, &(mnt->hdr.nrows))) {
	return(FALSE);
    }
    if (!xdr_u_int(xdrsp, &(mnt->hdr.ncols))) {
	return(FALSE);
    }
    if (!xdr_enum(xdrsp, (enum_t *)&(mnt->hdr.storageType))) {
	return(FALSE);
    }
    if (!xdr_array(xdrsp, (char **)&(mnt->points),
		   &size, size, sizeof(GDHE_MNT_POINT), xdr_mnt_point)) {
	return(FALSE);
    }
    return(TRUE);
}
#endif /* HAVE_GDHE_TERRAIN */

int
xdr_timeval(XDR *xdrsp, struct timeval *tv)
{
    if (!xdr_long(xdrsp, &(tv->tv_sec))) return FALSE;
    if (!xdr_long(xdrsp, &(tv->tv_usec))) return FALSE;
    return TRUE;
}

/*----------------------------------------------------------------------
  function name : xdr_packet
  original date : ??/??/1992
  modified dates: 03/05/1994  add 'select_robot_type'
------------------------------------------------------------------------
 * xdr_packet has a special status. We could make a call to *\
** xdr_union, but then we need define an enum type for the  **
** tag, and also i-don't-know-what. It is simpler to do the **
\* switch() ourselves.                                      *
----------------------------------------------------------------------*/
int xdr_packet ( XDR* xdrsp, packet* packetp )
{
  if ( !xdr_int ( xdrsp, &(packetp->command) ) )
    return FALSE;
  if ( !xdr_int ( xdrsp, &(packetp->which) ) )
    return FALSE;
  switch ( packetp->command )
    {
    case CLOSE_ALL:
    case DISCONNECT:
      break;
    case SET_PLATFORM:
    case MOVE_PLATFORM:
      if ( !xdr_reference ( xdrsp, (char**) &packetp->args.ang,
			   sizeof ( float ),
			   (xdrproc_t)xdr_float ) )
	  return FALSE;
      break;
    case PLACE_GOAL_O:
    case PLACE_GOAL_NO:
    case PLACE_ROBOT:
      if ( !xdr_reference ( xdrsp, ( char** ) &(packetp->args.pos),
			   sizeof ( position ),
			   (xdrproc_t)xdr_position ) )
	return FALSE;
      break;
    case PLACE_DIAMOND:
      if ( !xdr_reference ( xdrsp, ( char** ) &(packetp->args.vpos),
			   sizeof ( v_position ),
			   (xdrproc_t)xdr_v_position ) )
	return FALSE;
      break;
    case PLACE_DOORS:
    case PLACE_WALLS:
    case PLACE_WALLS_REL:
    case PLACE_WALLS_ACC:
      if ( !xdr_reference ( xdrsp, ( char** ) &(packetp->args.vseg),
			   sizeof ( v_segment ),
			   (xdrproc_t)xdr_v_segment ) )
	return FALSE;
      break;
    case PLACE_TRAJECTORY:
    case PLACE_PATH:
    case PLACE_ROBOT_TRAJ:
    case PLACE_ROBOT_PATH:
    case EXEC_TRAJ:
      
      if ( !xdr_reference ( xdrsp, ( char** ) &(packetp->args.traj),
			   sizeof ( trajectory ),
			   (xdrproc_t)xdr_trajectory ) )
	return FALSE;
      break;

    case PLACE_RS_TRAJ:
	if (xdr_reference(xdrsp, (char **)&(packetp->args.rs_traj),
			  sizeof(rs_trajectory), 
			  (xdrproc_t)xdr_rs_trajectory) == FALSE) {
	    return(FALSE);
	}
      break;
      
    case PLACE_FREE_SPACE:
      if ( !xdr_reference ( xdrsp, ( char** ) &(packetp->args.pol),
			   sizeof ( polygon ),
			   (xdrproc_t)xdr_polygon ) )
	return FALSE;
      break;
    case SELECT_ROBOT_TYPE:
      if ( !xdr_reference ( xdrsp, ( char** ) &(packetp->args.robot_type),
			   sizeof ( int ),
			   (xdrproc_t)xdr_int ) )
	return FALSE;
      break;
      case UPDATE_PER_OBST:
	if ( !xdr_reference ( xdrsp, ( char** ) &(packetp->args.pol),
			      sizeof ( polygon ),
			      (xdrproc_t)xdr_polygon ) )
	    return FALSE;
	break;
      case PLACE_DEPENDENCY_EDGE:
      case ERASE_DEPENDENCY_EDGE:
	if (!xdr_reference(xdrsp, (char **)&(packetp->args.edge_to),
			   sizeof (int),
			   (xdrproc_t)xdr_int)) {
	    return FALSE;
	}
	break;
      case PLACE_DEPENDENCY_CYCLE:
	if (!xdr_reference(xdrsp, (char **)&(packetp->args.cycle),
			   sizeof(dependency_cycle),
			   (xdrproc_t)xdr_dependency_cycle)) {
	    return FALSE;
	}
	break;
      case PLACE_MISSION_GOAL:
      case EVAL_EXPRESSION:
	if (!xdr_string(xdrsp, &(packetp->args.str),
			MISSION_GOAL_LENGTH)) {
	    return FALSE;
	}
	break;
#ifdef HAVE_GDHE_TERRAIN
      case PLACE_TERRAIN:
	if (!xdr_reference(xdrsp, (char **)&(packetp->args.terrain),
			   sizeof(GDHE_MNT), (xdrproc_t)xdr_mnt)) {
	    return FALSE;
	}
	break;
#endif /* HAVE_GDHE_TERRAIN */
	
    case ERASE_ROBOT:
    case ERASE_DIAMOND:
    case ERASE_DOORS:
    case ERASE_WALLS:
    case ERASE_FREE_SPACE:
    case ERASE_TRAJECTORY:
    case ERASE_PATH:
    case ERASE_ROBOT_TRAJ:
    case ERASE_ROBOT_PATH:
    case ERASE_RS_TRAJ:
    case ERASE_GOAL:
    case ERASE_PER_OBST:
    case VIDEO_ON:
    case LASER_ON:
    case VIDEO_OFF:
      case LASER_OFF:
      case ERASE_DEPENDENCY_CYCLE:
      case ERASE_MISSION_GOAL:
#ifdef HAVE_GDHE_TERRAIN
      case ERASE_TERRAIN:
#endif
      break;
    default:
      return FALSE;
    }
  return TRUE;
}
