/*
 * Copyright (C) 1996-2001 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*-
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Revision History:
 * 9-Oct-98:  dek@cgl.ucsf.edu  Added stars.
 *
 * 8-Oct-98:  jwz@jwz.org   Made the 512x512x1 xearth image be built in.
 *                          Made it possible to load XPM or XBM files.
 *                          Made the planet bounce and roll around.
 *
 * 8-Oct-98: Released initial version of "glplanet"
 * (David Konerding, dek@cgl.ucsf.edu)
 *
 * BUGS:
 * -bounce is broken
 * 
 *   For even more spectacular results, grab the images from the "SSysten"
 *   package (http://www.msu.edu/user/kamelkev/) and do this:
 *
 *     cd ssystem-1.4/hires/
 *     foreach f ( *.jpg )
 *       djpeg $f | ppmquant 254 | ppmtoxpm > /tmp/$f:r.xpm
 *     end
 *
 *     cd /tmp
 *     foreach f ( *.xpm )
 *       glplanet -image $f
 *     end
 */



#include <X11/xpm.h>
#ifndef PIXEL_ALREADY_TYPEDEFED
#  define PIXEL_ALREADY_TYPEDEFED /* Sigh, Xmu/Drawing.h needs this... */
#endif
#include <X11/Xmu/Drawing.h>
#include <GL/glu.h>

#include <stdlib.h>
#include <math.h>
#include <string.h>

#include <tcl.h>
#define TOGL_X11
#include <togl.h>

static char *which_image = NULL;


#include "../../images/earth.xbm"
#include "xpm-ximage.h"

#ifndef __GNUC__
#define inline 
#endif

/*-
 * slices and stacks are used in the sphere parameterization routine.
 * more slices and stacks will increase the quality of the sphere,
 * at the expense of rendering speed
 */

#define SLICES 24
#define STACKS 24

/* radius of the sphere- fairly arbitrary */
#define RADIUS 10

#define BG_COLOR "blue"
#define FG_COLOR "green"

/* structure for holding the planet data */
typedef struct {
  GLuint platelist;
  XColor fg, bg;
  int orientation;
} planetstruct;


static planetstruct planet;


static inline void
normalize(GLfloat v[3])
{
	GLfloat     d = (GLfloat) sqrt((double) (v[0] * v[0] + v[1] * v[1] + v[2] * v[2]));

	if (d != 0) {
		v[0] /= d;
		v[1] /= d;
		v[2] /= d;
	} else {
		v[0] = v[1] = v[2] = 0;
	}
}


/* Set up and enable texturing on our object */
static void
setup_xbm_texture (unsigned char *bits, int width, int height,
				   XColor *fgc, XColor *bgc)
{
  unsigned int fg = (((fgc->red  >> 8) << 16) |
					 ((fgc->green >> 8) << 8) |
					 ((fgc->blue >> 8)));
  unsigned int bg = (((bgc->red  >> 8) << 16) |
					 ((bgc->green >> 8) << 8) |
					 ((bgc->blue >> 8)));

  unsigned char *data = (unsigned char *)
	malloc ((width * height * 24) / 8);
  unsigned char *out = data;
  int x, y;

  for (y = 0; y < height; y++)
	for (x = 0; x < width; x++)
	  {
		unsigned char byte = bits [(y * (width / 8) + (x / 8))];
		unsigned char bit = (byte & (1 << (x % 8))) >> (x % 8);
		unsigned int word = (bit ? bg : fg);
		*out++ = (word & 0xFF0000) >> 16;
		*out++ = (word & 0x00FF00) >> 8;
		*out++ = (word & 0x0000FF);
	  }

  glTexImage2D(GL_TEXTURE_2D, 0, 3, width, height, 0,
			   GL_RGB, GL_UNSIGNED_BYTE, data);

  /* setup parameters for texturing */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
}


static void
setup_file_texture (struct Togl *togl, char *filename)
{
    Display *dpy = Togl_Display(togl);
    Visual *visual = Togl_Visual(togl);
    Colormap cmap = Togl_Colormap(togl);

    char **xpm_data = 0;
    planetstruct *gp = &planet;
    unsigned int width = 0;
    unsigned int height = 0;
    unsigned char *data = 0;
    int xhot, yhot;
    int result;

    result = XpmReadFileToData (filename, &xpm_data);

    switch (result) {
      case XpmSuccess:
      {
	  XImage *image = xpm_to_ximage (dpy, visual, cmap, xpm_data);
	  
	  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
		       image->width, image->height, 0,
		       GL_RGBA, GL_UNSIGNED_BYTE, image->data);
	  
	  /* setup parameters for texturing */
	  glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
	  glPixelStorei(GL_UNPACK_ROW_LENGTH, image->width);
	  
	  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	  return;
      }
      /*NOTREACHED*/
      break;
      
      case XpmOpenFailed:
	fprintf (stderr, "gdhe: file %s doesn't exist.\n",  filename);
	exit (-1);
	break;
	
      case XpmFileInvalid:
	/* Fall through and try it as an XBM. */
	break;
	
      case XpmNoMemory:
	fprintf (stderr, "gdhe: XPM: out of memory\n");
	exit (-1);
	break;
	
      default:
	fprintf (stderr, "gdhe: XPM: unknown error code %d\n",  result);
	exit (-1);
	break;
    } /* switch */


    result = XmuReadBitmapDataFromFile (filename, &width, &height, &data,
					    &xhot, &yhot);
    if (result != Success)
    {
	fprintf (stderr, "gdhe: not an XBM file: %s\n", filename);
	exit (1);
    }
    setup_xbm_texture (data, width, height, &gp->fg, &gp->bg);
    
}


static void
setup_texture(struct Togl *togl)
{
  planetstruct *gp = &planet;
  if (!which_image ||
	  !*which_image ||
	  !strcmp(which_image, "BUILTIN"))
	setup_xbm_texture (earth_bits, earth_width, earth_height,
					   &gp->fg, &gp->bg);
  else
	setup_file_texture (togl, which_image);
}



/* Set up and enable face culling so we don't see the inside of the sphere */
static void
setup_face(int orientation)
{
  glEnable(GL_CULL_FACE);
  if (orientation > 0) {
      glCullFace(GL_BACK); 
  } else {
      glCullFace(GL_FRONT);
  }
}


/* Function for determining points on the surface of the sphere */
static void inline 
ParametricSphere(float theta, float rho, GLfloat *vector)
{
  vector[0] = -sin(theta) * sin(rho);
  vector[1] = cos(theta) * sin(rho);
  vector[2] = cos(rho);

#if DO_HELIX
  vector[0] = -(1- cos(theta)) * cos(rho); 
  vector[1] = -(1- cos(theta)) * sin(rho); 
  vector[2] = -(sin(theta) + rho); 
#endif /* DO_HELIX */

	return;
}


/* Initialization function  */
static void
pinit(struct Togl *togl, double radius, int orientation)
{
    planetstruct *gp = &planet;
    int i, j;
    int stacks=STACKS, slices=SLICES;
    Display *dpy = Togl_Display(togl);
    Colormap cmap = Togl_Colormap(togl);
    float drho, dtheta;
    float rho, theta;
    GLfloat vector[3];
    GLfloat ds, dt, t, s;;
  
  
    if (!XParseColor(dpy, cmap, FG_COLOR, &gp->fg)) {
	fprintf(stderr, "unsparseable color: \"%s\"\n", FG_COLOR);
    }
    if (!XParseColor(dpy, cmap, BG_COLOR, &gp->bg)) {
	fprintf(stderr, "unsparseable color: \"%s\"\n", BG_COLOR);
    }
    
    setup_texture(togl);
    
    setup_face(orientation);
    gp->orientation = orientation;

    /*-
     * Generate a sphere with quadrilaterals.
     * Quad vertices are determined using a parametric sphere function.
     * For fun, you could generate practically any parameteric surface and
     * map an image onto it. 
     */
    
    drho = M_PI / stacks;
    dtheta = 2.0 * M_PI / slices;
    ds = 1.0 / slices;
    dt = 1.0 / stacks;
    
    
    gp->platelist=glGenLists(1);
    glNewList(gp->platelist, GL_COMPILE);
    
    glColor3f(1,1,1);
    glBegin(GL_QUADS);
    
    t = 0.0;
    for(i=0; i<stacks; i++) {
	rho = i * drho;
	s = 0.0;
	for(j=0; j<slices; j++) {
	    theta = j * dtheta;
	    
	    
	    glTexCoord2f(s,t);
	    ParametricSphere(theta, rho, vector);
	    normalize(vector);
	    glNormal3fv(vector);
	    ParametricSphere(theta, rho, vector);
	    glVertex3f( vector[0]*radius, vector[1]*radius, vector[2]*radius );
	    
	    glTexCoord2f(s,t+dt);
	    ParametricSphere(theta, rho+drho, vector);
	    normalize(vector);
	    glNormal3fv(vector);
	    ParametricSphere(theta, rho+drho, vector);
	    glVertex3f( vector[0]*radius, vector[1]*radius, vector[2]*radius );
	    
	    glTexCoord2f(s+ds,t+dt);
	    ParametricSphere(theta + dtheta, rho+drho, vector);
	    normalize(vector);
	    glNormal3fv(vector);
	    ParametricSphere(theta + dtheta, rho+drho, vector);
	    glVertex3f( vector[0]*radius, vector[1]*radius, vector[2]*radius );
	    
	    glTexCoord2f(s+ds, t);
	    ParametricSphere(theta + dtheta, rho, vector);
	    normalize(vector);
	    glNormal3fv(vector);
	    ParametricSphere(theta + dtheta, rho, vector);
	    glVertex3f( vector[0]*radius, vector[1]*radius, vector[2]*radius );

	    s = s + ds;
	    
	}
	t = t + dt;
    }
    glEnd();
    glEndList();


}

/*----------------------------------------------------------------------*/

static void
draw_sphere(void)
{
    planetstruct *gp = &planet;
    
    setup_face(gp->orientation);
    glEnable(GL_DEPTH_TEST);
    
    /* turn on the various attributes for making the sphere look nice */
    glEnable(GL_TEXTURE_2D);
    
    glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
    glEnable(GL_COLOR_MATERIAL);
    
    glCallList(gp->platelist);
}






void
draw_planet(void)
{
    
    /* protect our modelview matrix and attributes */
    glPushMatrix();
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    /* draw the sphere */
    draw_sphere();
    
    glPopMatrix();
    glPopAttrib();
}


/*----------------------------------------------------------------------*/

int
initPlanetCallBack(struct Togl *togl, int argc, char *argv[])
{
    int orientation = 1;
    double radius = 10.;
    int error = 0;
    Tcl_Interp *interp = Togl_Interp(togl);

    /* Grr, il faudrait vraiment un getopt_long */
    while (argc > 3) {
	if (!strcmp(argv[2], "-interior")) {
	    orientation = -1;
	} else 	if (!strcmp(argv[2], "-radius")) {
	    if (argc > 3) {
		radius = atof(argv[3]);
		argc--;
		argv++;
	    } else {
		error++;
	    }
	} else {
	    error++;
	}
	argc--;
	argv++;
    }
	
    if (!error && argc == 3) {
	which_image = argv[2];
    } else {
	Tcl_SetObjResult(interp, 
			 Tcl_NewStringObj("syntax: widget \"initPlanet\""
					  "[-interior] filename", -1));
	return TCL_ERROR;
    }
    /* protect our modelview matrix and attributes */
    glPushMatrix();
    pinit(togl, radius, orientation);
    glPopMatrix();
    
    interp->result[0] = '\0';
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

int
drawPlanetCallBack(ClientData clientData, Tcl_Interp *interp, 
		   int argc, Tcl_Obj *const argv[])
{
    draw_planet();
    return TCL_OK;
}
/*----------------------------------------------------------------------*/

int
Planet_Init(Tcl_Interp *interp)
{
    int code;

    code = Tcl_PkgProvide(interp, "Planet", "1.0");
    if (code != TCL_OK) {
	return code;
    }

    Togl_CreateCommand("initPlanet", initPlanetCallBack);
    Tcl_CreateObjCommand(interp, "drawPlanet", drawPlanetCallBack, NULL, NULL);
    /* Let result string empty */
    Tcl_SetObjResult(interp, Tcl_NewStringObj("", -1));
    return TCL_OK;
}
